#! /usr/bin/env python
# -*- coding: utf-8 -*-

#
# このコードは Pillow と pywin32 を使用しています。
#
# Pillowのインストールはコマンドプロンプトで以下のどちらかを実行
# py -m pip install Pillow
# pip install Pillow
#
# pywin32のインストールはコマンドプロンプトで以下のどちらかを実行
# py -m pip install pywin32
# pip install pywin32
#
# Pillowは、開発が停止しているPIL(Python Image Library)からフォークされた
# 画像処理ライブラリ。
# PillowとPILは共存出来ない。
# インストールするのはPillowだがインポートするパッケージの名前は
# PillowではなくPILなので注意。
#
# Windowsでクリップボードを利用するには、
# pywin32パッケージに含まれているwin32clipboardモジュールを使う
# これはWindows専用で、Linuxなどでは使えないらしい。
#

import os, io, sys, time, threading
import tkinter, tkinter.ttk, tkinter.messagebox, tkinter.filedialog
from PIL import Image, ImageTk, ImageDraw
import win32clipboard
import Maze

# Windowsのクリップボード用の定番コード
# メモリストリームにBMP形式で保存してから読み出す事で、
# Windows用のビットマップ形式を得る。
# CF_DIB(8)はWIndowsでクリップボードのビットマップ形式を指定する値
def copy_to_clipboard(original_image):
    output = io.BytesIO()
    original_image.convert('RGB').save(output, 'BMP')
    data = output.getvalue()[14:]   # 14はBITMAPFILEHEADERのサイズ。
    output.close()
    send_to_clibboard(win32clipboard.CF_DIB, data)

# Windowsのクリップボードへコピー
def send_to_clibboard(clip_type, data):
    win32clipboard.OpenClipboard()
    win32clipboard.EmptyClipboard()
    win32clipboard.SetClipboardData(clip_type, data)
    win32clipboard.CloseClipboard()

# 「ゆっくり作る」の実装のため、迷路は別スレッドで作成する。
# ウィンドウサイズを変更する際に、何度も迷路を作り直そうとするので、
# 無駄な作成をしないために、迷路を作り始めるまで0.2秒待って作成開始
class MakeMazeThread(threading.Thread):
    # コンストラクタ
    def __init__(self, window, timer):
        super().__init__()
        self.window = window
        self.timer = timer
        self.forcedStop= False
    # 強制停止フラグの設定
    # スレッドを無理に止める方法がわからなかった。
    def stop(self):
        self.forcedStop = True
    def run(self):
        time.sleep(0.2) # 0.2秒待つ
        if self.forcedStop: return
        # 迷路を生成して描画
        self.window.makeMazeInit()
        if self.forcedStop: return
        while self.window.makeMazeStep():
            if self.forcedStop: return
            if(self.timer > 0):
                self.window.makeMazeDraw()
                time.sleep(self.timer)
            if self.forcedStop: return
        if self.forcedStop: return
        self.window.makeMazeDraw()

# GUI Windowを作成してアプリケーションを作る。
class MazeWindow(tkinter.Tk):
    # 迷路生成の初期化
    # これはスレッドから呼ばれる。
    # canvasの大きさを基準に、表示できるように迷路の縦横サイズを決定する。
    # 迷路作成にそこそこ時間がかかるので、画面の下に進捗バーを表示する。
    def makeMazeInit(self):
        d = self.dotsize.get()
        w = self.canvas.winfo_width()
        h = self.canvas.winfo_height()
        x = (w-5) // d
        y = (h-5) // d
        s = 20 // d
        self.maze = Maze.MakeMaze(x, y, s, 1, d)
        self.status.configure(maximum=self.maze.getFinish(), value=0)
        self.update()
    # 1段階の迷路作成
    # これはスレッドから呼ばれる。
    # 迷路が完成したら、進捗バーを消す。
    def makeMazeStep(self):
        if self.maze.make():
            self.status.configure(value=self.maze.getProgress())
            return True
        else:
            self.status.configure(value=0)
            return False
    # 迷路クラスの表示
    # これはスレッドから呼ばれる。
    # チラつき防止のため、新図形描画後に旧図形を削除する。
    def makeMazeDraw(self):
        # 旧図形保護
        old1 = self.mazeImage
        old2 = self.mazePhoto
        old3 = self.imageID
        # 迷路描写、キャンバスの中心に設定
        self.mazeImage = self.maze.toImage()
        self.mazePhoto = ImageTk.PhotoImage(self.mazeImage)
        w = self.canvas.winfo_width()
        h = self.canvas.winfo_height()
        self.imageID = self.canvas.create_image(w//2, h//2, image = self.mazePhoto, tags = 'maze')
        # 古い迷路の図を図形を削除
        if old3 is not None:
            self.canvas.delete(old3)
    
    
    # 大きさ変更チェック
    # ウィンドウを移動した際も<Configure>イベントが発生するので、
    # 無駄に迷路を作り直さないようにする。
    def isResize(self):
        if self.lastX != self.canvas.winfo_width():
            self.lastX = self.canvas.winfo_width()
            self.lastY = self.canvas.winfo_height()
            return True
        if self.lastY != self.canvas.winfo_height():
            self.lastX = self.canvas.winfo_width()
            self.lastY = self.canvas.winfo_height()
            return True
        return False
    # 迷路生成。
    # ユーザのメニュー選択やショートカットなどから呼ばれる
    def makeMaze(self, event=None):
        self.thred.stop()
        self.thred = MakeMazeThread(self, 0)
        self.thred.start()
    # 大きさ変更
    # <Configure>イベントから呼ばれる
    def resize(self, event=None):
        if not self.isResize(): return
        self.makeMaze()
    # ゆっくり作る
    # ユーザのメニュー選択から呼ばれる
    def makeSlow(self, event=None):
        self.thred.stop()
        self.thred = MakeMazeThread(self, 0.05)
        self.thred.start()
    # 解答を生成
    # ユーザのメニュー選択やショートカットから呼ばれる
    # 迷路生成中は答えを作れないので、メッセージを表示する。
    def makeAnswer(self, event=None):
        if self.thred.is_alive():
            tkinter.messagebox.showwarning('注意', '迷路生成中に解答を作成できません')
            return
        self.maze.answer()
        self.makeMazeDraw()
    # 終了処理
    # ユーザのメニュー選択やショートカットから呼ばれる
    # スレッドを停止し、1秒後にアプリケーション停止を設定
    # スレッドがすぐ停止するとは限らないので1秒待つ。
    # join()を使うとうまくいかなかった。理由は知らない。
    def finish(self, event=None):
        self.thred.stop()
        self.after(1000, self.destroy)
    # 画像をクリップボードに入れる
    # 迷路生成中とか気にせずクリップボードに入れる
    def cut(self, event=None):
        copy_to_clipboard(self.mazeImage)
    # 画像をファイル保存
    # 迷路生成中とか気にせず保存する
    # defaultextensionを設定すると変な拡張子にならない様だ。
    # save()は拡張子から保存するファイル形式を自動判断する様だ。
    def save(self, event=None):
        fileType = [('PNG','.png'),('JPEG','.jpg'),('GIF','.gif'),('BMP','.bmp')]
        file = tkinter.filedialog.asksaveasfilename(
            defaultextension='.png', filetypes = fileType, title='保存ファイル名を入れてください')
        if not file: return
        try:
            self.mazeImage.save(file)
        except ValueError as e:  # unknown file extension
            tkinter.messagebox.showerror('エラー', str(e))
    
    
    # MazeWindowクラスのコンストラクタ
    # ウィンドウサイズを調整したり、メニューを作成したり、
    # 迷路の描画領域を作ったりといろいろ。
    def __init__(self, master = None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        #ウィンドウを640x400にする
        self.geometry("640x400")
        # 最小ウィンドウサイズ
        self.minsize(100, 100)
        #ウィンドウのタイトルを変更
        self.title("迷路プログラム１")
        #メニューバー作成
        self.dotsize = tkinter.IntVar()
        self.dotsize.set(20)
        menubar = tkinter.Menu(self)
        self.configure(menu = menubar)
        # メニューバー
        filemenu = tkinter.Menu(menubar, tearoff=False)
        editmenu = tkinter.Menu(menubar, tearoff=False)
        sizemenu = tkinter.Menu(menubar, tearoff=False)
        menubar.add_cascade(label="ファイル(F)", underline = 5, menu=filemenu)
        menubar.add_cascade(label="編集(E)",     underline = 3, menu=editmenu)
        menubar.add_cascade(label="サイズ(S)",   underline = 4, menu=sizemenu)
        # ファイルメニュー
        filemenu.add_command(label="保存", accelerator='Ctrl+S', command=self.save)
        filemenu.add_command(label="終了", accelerator='Ctrl+Q', command=self.finish)
        self.bind_all("<Control-s>", self.save)
        self.bind_all("<Control-q>", self.finish)
        # 編集メニュー
        editmenu.add_command(label="新規", accelerator='Ctrl+N', command=self.makeMaze)
        editmenu.add_command(label="ゆっくり作る",               command=self.makeSlow)
        editmenu.add_command(label="解答", accelerator='Ctrl+A', command=self.makeAnswer)
        editmenu.add_command(label="コピー", accelerator='Ctrl+C', command=self.cut)
        self.bind_all("<Control-n>", self.makeMaze)
        self.bind_all("<Control-a>", self.makeAnswer)
        self.bind_all("<Control-c>", self.cut)
        # サイズメニュー
        sizemenu.add_radiobutton(label="大", variable=self.dotsize, value=5 , command=self.makeMaze)
        sizemenu.add_radiobutton(label="中", variable=self.dotsize, value=10, command=self.makeMaze)
        sizemenu.add_radiobutton(label="小", variable=self.dotsize, value=20, command=self.makeMaze)
        # キャンバス
        self.canvas = tkinter.Canvas(self, bg = 'white')
        self.canvas.grid(row=0, column=0, columnspan=2, \
                         sticky=(tkinter.W, tkinter.E, tkinter.N, tkinter.S) )
        # プログレスバーの設定
        self.status = tkinter.ttk.Progressbar(self, orient='horizontal', mode='determinate')
        self.status.grid(row=1, column=0, sticky=(tkinter.W, tkinter.E))
        # ウィンドウがリサイズ可能であることを明示的に示すサイズグリップ
        self.sizegrip = tkinter.ttk.Sizegrip(self)
        self.sizegrip.grid(row=1, column=1, sticky=(tkinter.S, tkinter.E))
        # 拡大縮小設定
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        # 大きさ変更検知用
        self.lastX = 0
        self.lastY = 0
        # 大きさ変更後に再作成用
        self.canvas.bind('<Configure>', self.resize)
        # 迷路作成スレッド
        self.thred = MakeMazeThread(self, 0)
        self.mazeImage = None
        self.mazePhoto = None
        self.imageID = None

# メインルーチン、ウィンドを作成して実行
def main():
    root = MazeWindow()
    root.mainloop()

# 直接呼ばれたら、main()を実行する。
if __name__ == '__main__':
    main()

