#! /usr/bin/env python
# -*- coding: utf-8 -*-

#
# このコードは Pillow を使用しています。
#
# Pillowのインストールはコマンドプロンプトで以下のどちらかを実行
# py -m pip install Pillow
# pip install Pillow
#
# Pillowは、開発が停止しているPIL(Python Image Library)からフォークされた
# 画像処理ライブラリ。
# PillowとPILは共存出来ない。
# インストールするのはPillowだがインポートするパッケージの名前は
# PillowではなくPILなので注意。
#
# Pythonでモジュールを使う場合に普通は import hogehohe でいいのですが、
# Image, ImageDraw は import PIL ではだめで from PIL としなければ
# なりませんでした。よくわかりません。
#

import os, random
from PIL import Image, ImageDraw

# 迷路を描くためのボード
class MazeBoard:
    # コンストラクタ。迷路サイズはインスタンス生成時に決まり、変更できない。
    #
    # Cはメモリ管理が面倒なので、１回確保したメモリの再利用をしていました。
    # オブジェクト型言語はオブジェクトの使い捨てが簡単なため、
    # 迷路を作成する毎に新しいオブジェクトを作る事にしました。
    #
    # 四方の壁はコンストラクタで作ってます。
    # スタートは左上、ゴールは右下固定で、コンストラクタで作っています。
    #
    # 迷路の基礎部分は縦横に線が書ける方眼紙の様な２次元配列（list in list）です。
    # 迷路サイズの縦横２倍の配列を作り、壁の部分と通路の部分を同じ配列に並べています。
    # 迷路サイズが小さすぎる場合はエラーとします。
    # 迷路の基礎部分のデータ構造をどうするかは、いろいろな選択肢があるので、
    # getXXX() や setXXX() といった関数で隠蔽しています
    #
    def __init__(self, XSize, YSize, startSize, lineWidth=1, cellSize=0):
        # 引数が妥当な値であるかのチェック
        self.error = None
        if startSize < 1:
            self.error = "スタートが小さすぎ({})".format(startSize)
            startSize = 1
        if cellSize == 0: cellSize = 20//startSize
        if cellSize < lineWidth*3:
            self.error = "スタートが大きすぎ({})".format(startSize)
            cellSize = lineWidth*3
        if XSize < startSize*2+2:
            self.error = "横幅が小さすぎ({}<{})".format(XSize, startSize*2+2)
            XSize = startSize*2+2
        if YSize < startSize*2+2:
            self.error = "縦幅が小さすぎ({}<{})".format(YSize, startSize*2+2)
            YSize = startSize*2+2
        # 引数などをインスタンス変数に保存
        self.XSize = XSize
        self.YSize = YSize
        self.SSize = startSize
        self.LSize = lineWidth
        self.CSize = cellSize
        self.ImgXSize = XSize * cellSize + lineWidth * 3
        self.ImgYSize = YSize * cellSize + lineWidth * 3
        self.mode = 0
        if self.error is not None:
            return
        # 迷路形状を示す２次元配列
        self.Data = [[0 for i in range(self.YSize*2+1)] for j in range(self.XSize*2+1)]
        # 四方の壁を作る
        for i in range(self.YSize*2):
            self.Data[0][i] = 1
            self.Data[self.XSize*2][i] = 1
        for j in range(self.XSize*2):
            self.Data[j][0] = 1
            self.Data[j][self.YSize*2] = 1
        # スタートとゴールを作る
        for x in range(self.SSize*2):
            for y in range(self.SSize*2):
                self.Data[x][y] = 2
                self.Data[self.XSize*2-x][self.YSize*2-y] = 3
    
    # 壁描画モードと、通路探索モードの切り替え
    def wallMode(self):
        self.mode = 0
    def routeMode(self):
        self.mode = 1
    
    # インスタンス変数の読み書き
    def getXSize(self):
        return self.XSize + 1 - self.mode
    def getYSize(self):
        return self.YSize + 1 - self.mode
    def getStartSize(self):
        return self.SSize
    def getLineWidth(self):
        return self.LSize
    def getCellSize(self):
        return self.CSize
    def getImgXSize(self):
        return self.ImgXSize
    def getImgYSize(self):
        return self.ImgYSize
    
    # 配列のインデックスチェックは最大値と比較すればいいのですが、
    # エラーを発生させてキャッチでもいいので、 try:...except IndexError: を使っています。
    # Pythonは配列のインデックスがマイナスの場合、エラーではなく後ろからの順番になります。
    # このためマイナスでない事のチェックは必要でした。
    def checkMinus(self, v):
        if v < 0 : raise IndexError
    
    # 迷路の設定や読み出し
    def walk(self, x, y, d):
        if d == 1:
            return x+1, y
        elif d == 2:
            return x, y+1
        elif d == 3:
            return x-1, y
        elif d == 4:
            return x, y-1
        else:
            raise Exception("方向指定バグ")
    def getData(self, x, y, d = 0):
        if self.error is not None: return 1
        if d != 0:
            x, y = self.walk(x, y, d)
        try:
            self.checkMinus(x)
            self.checkMinus(y)
            return self.Data[x*2+self.mode][y*2+self.mode]
        except IndexError:
            return 1
    def setData(self, x, y, v):
        if self.error is not None: return
        try:
            self.checkMinus(x)
            self.checkMinus(y)
            self.Data[x*2+self.mode][y*2+self.mode] = v
        except IndexError:
            pass
    def getWall(self, x, y, d):
        if self.error is not None: return 1
        x, y = self.walk(x*2+self.mode, y*2+self.mode, d)
        try:
            self.checkMinus(x)
            self.checkMinus(y)
            return self.Data[x][y]
        except IndexError:
            return 1
    def setWall(self, x, y, d, v):
        if self.error is not None: return
        x, y = self.walk(x*2+self.mode, y*2+self.mode, d)
        try:
            self.checkMinus(x)
            self.checkMinus(y)
            self.Data[x][y] = v
        except IndexError:
            pass
    
    # 迷路を文字列で表現
    # デバッグ時には役立つが、本番では不要かも
    def toString(self):
        if self.error is not None: return self.error
        r = ''
        for y in range(self.YSize*2+1):
            for x in range(self.XSize*2+1):
                if self.Data[x][y] == 1:
                    r += '+-| '[(y%2)*2+(x%2)]
                else:
                    r += ' *SG#'[ self.Data[x][y] ]
            r += '\n'
        return r
    
    # 迷路をビットマップ画像で表現
    # ビットマップはクリップボードで他のアプリに渡したり、
    # ファイルに保存できて便利
    # スタートとゴールは19x19ドットのPNGの図を読み込んでいます。
    # PNGファイルはPython実行ファイルと同じフォルダにあると想定しています。
    # Pythonは __file__ 変数がPython実行ファイルを示すので、
    # このディレクトリ部分とファイル名を結合していファイルパスを作成して、
    # ファイルを読んでいます。
    # brd.pasteの３つ目のパラメータは画像を張り込む際のマスクです。
    # PNG画像は４面からなり、４番目が透明度になります。
    # 画像を分割した４番目の面( .split()[3] )の透明度をマスクに使用します。
    def toImage(self):
        # 白い土台
        brd = Image.new('RGB', (self.ImgXSize, self.ImgYSize) , 'white')
        if self.error is not None: return brd
        # 図形に書き込みをするためのオブジェクト
        drw = ImageDraw.Draw(brd)
        # 周りに線１本分の空白を作る
        offset = self.LSize
        # 解答作成前と作成後でスタートとゴールの色を変える
        # Data[1][1]を参照するのは美しくないが、まあいいか
        goleColor = 'lightgreen' if self.Data[1][1] > 3 else 'white'
        colorTable = ['white','black',goleColor,goleColor,'lightgreen']
        # 壁などの描画。
        for y in range(self.YSize*2+1):
            if (y%2) == 0:
                y1 = y/2 * self.CSize + offset
                y2 = y1 + self.LSize
            else:
                y1 = y2
                y2 = (y+1)/2 * self.CSize + offset
            for x in range(self.XSize*2+1):
                if (x%2) == 0:
                    x1 = x/2 * self.CSize + offset
                    x2 = x1 + self.LSize
                else:
                    x1 = x2
                    x2 = (x+1)/2 * self.CSize + offset
                c = colorTable[ self.Data[x][y] ]
                drw.rectangle((x1, y1, x2-1, y2-1), fill=c)
        # スタートとゴールの描写
        # pngファイルは、実行ファイルと同じ場所にあると仮定
        # 実行ディレクトリと実行ファイルの場所は違うかもしれない。
        # __file__が実行パスなので、これのディレクトリ名とファイル名を結合する。
        stt = Image.open(os.path.join(os.path.dirname(__file__),'start.png'))
        brd.paste(stt, ( offset, offset ), stt.split()[3])
        gol = Image.open(os.path.join(os.path.dirname(__file__),'goal.png'))
        brd.paste(gol, ( brd.width - gol.width - offset \
                      , brd.height - gol.height - offset ), gol.split()[3])
        return brd

# 迷路を作るコード
class MakeMaze(MazeBoard):
    # コンストラクタ
    def __init__(self, XSize, YSize, StartSize, LineWidth=1, CellSize=0):
        super().__init__(XSize, YSize, StartSize, LineWidth, CellSize)
        # 壁位置を覚えるためのスタック
        # CはPushやPopをするスタックライブラリがなかったので、スタックを
        # 自作しましたが、Pythonはlistがスタックとしても使えるので、楽です
        self.stack = list()
        # 各種イスタンス変数設定
        self.status = 2
        self.currentX = 0
        self.currentY = 0
        xs = self.getXSize()
        ys = self.getYSize()
        # 進捗を測るための変数
        self.progress = 0
        self.finish = (xs-2) * (ys-2)
        # 壁位置をスタックに保存
        for i in range(xs):
            self.stack.append( ( i, 0) )
            self.stack.append( ( i, ys-1) )
        for j in range(ys):
            self.stack.append( ( 0, j) )
            self.stack.append( ( xs-1, j) )
        # スタックで周囲の壁が書かれている範囲
        self.mark = len(self.stack)
    
    # インスタンス変数の参照
    def getProgress(self):
        return self.progress
    def getFinish(self):
        return self.finish
    def getStep(self):
        return self.status
    
    # 迷路生成ルーチン
    # 一気に作るのではなく、何度も呼出される
    # 完成するとStatusが０になり、make関数はFalseを返す
    # 一気に作らない理由は、迷路生成過程をわかりやすく見せるための
    # 「ゆっくり作る」を実装するためです。
    def make(self):
        # 戦略その１
        if self.status == 2:
            if self.step2() == False:
                self.status = 3
            return True
        # 戦略その２
        elif self.status == 3:
            if self.step3() == False:
                self.status = 4
            return True
        # 最後の穴埋め
        elif self.status == 4:
            self.step4()
            self.stack = None
            self.status = 0
            return True
        # 完成
        else:
            return False
    
    # 周囲の移動可能性の検査
    def gocheck(self, x, y):
        r = [];
        if self.getData(x, y, 1) == 0:
            r.append(1)
        if self.getData(x, y, 2) == 0:
            r.append(2)
        if self.getData(x, y, 3) == 0:
            r.append(3)
        if self.getData(x, y, 4) == 0:
            r.append(4)
        return r
    # 移動方向のランダム決定
    def gorand(self, d):
        return d[ random.randrange(len(d)) ]
    # 壁を延ばす
    def extend(self, x, y, d):
        if self.getData(x, y, d) != 0: return x, y  # 壁に当たった
        self.setWall(x, y, d, 1)
        x, y = self.walk(x, y, d)
        self.setData(x, y, 1)
        self.stack.append( (x, y) )
        self.progress += 1
        return x, y
    
    # STEP2:長く壁を伸ばす戦略その１
    # 外壁をランダムに決定し、壁を伸ばす
    # １０回連続失敗したら、false を返す
    # 壁を伸ばせるなら、適度に直進を試みる
    # 壁を伸ばしたら位置をスタックへ積む
    def step2(self):
        x = self.currentX
        y = self.currentY
        # スタート地点を決定
        for i in range(10):
            d = self.gocheck(x, y)
            if len(d) > 0: break
            x, y = self.stack[ random.randrange(self.mark) ]
        if len(d) <= 0: return False
        # 移動方向と長さを決定
        c = self.gorand(d)
        if c==1 or c==3:
            n = random.randrange(self.getXSize() // 8 + 1) + 1
        elif c==2 or c==4:
            n = random.randrange(self.getXSize() // 8 + 1) + 1
        else:
            raise Exception("バグです。")
        # 壁を延ばす
        for i in range(n):
            x, y = self.extend(x, y, c)
        self.currentX = x
        self.currentY = y
        return True
    
    # STEP3:長く壁を伸ばす戦略その２
    # スタックからランダムに読み出して、壁を伸ばす
    # １００回連続失敗したら、false を返す
    # いける方向にランダムに壁を伸ばす
    def step3(self):
        x = self.currentX
        y = self.currentY
        # スタート地点を決定
        for i in range(100):
            d = self.gocheck(x, y)
            if len(d) > 0: break
            x, y = self.stack[ random.randrange(len(self.stack)) ]
        if len(d) <= 0: return False
        # 移動方向と長さを決定
        c = self.gorand(d)
        if c==1 or c==3:
            n = random.randrange(self.getXSize() // 10 + 1) + 1
        elif c==2 or c==4:
            n = random.randrange(self.getXSize() // 10 + 1) + 1
        else:
            raise Exception("バグです。")
        # 壁を延ばす
        for i in range(n):
            x, y = self.extend(x, y, c)
        self.currentX = x
        self.currentY = y       
        return True
    
    # STEP4 最後の穴埋め
    # スタックの後ろからデータを読み、そこから壁を延ばす。
    # スタックが空になるまで続ける。
    def step4(self):
        while len(self.stack) > 0:
            x, y = self.stack.pop()
            d = self.gocheck(x, y)
            if len(d)>0:
                c = self.gorand(d)
                self.extend(x, y, c)

    # 道の探索
    def answerSub1(self):
        # 探索済み位置をマークするためのボード
        work = MazeBoard(self.XSize, self.YSize, 1)
        work.routeMode()
        # スタート点
        x = 0;
        y = 0;
        work.setData(x, y, 4)
        while True:
            # 進む方向をチェック
            for d in range(1, 6):
                if d > 4:                           # 道無し
                    x, y, d = self.stack.pop()
                elif self.getWall(x, y, d) == 1:    # 壁に当たる
                    pass
                elif self.getData(x, y, d) == 3:    # ゴール
                    self.stack.append( (x, y, d) )
                    return
                elif work.getData(x, y, d) == 0:    # 行ける
                    self.stack.append( (x, y, d) )
                    x, y = self.walk(x, y, d)
                    work.setData(x, y, 4)
                    break
                else:                               # 探索済み
                    pass
    # GOAL処理;
    # ポップを繰り返して道を塗る
    def answerSub2(self):
        while len(self.stack) > 0:
            x, y, d = self.stack.pop()
            self.setWall(x, y, d, 4)
            self.setData(x, y, 4)
    # 解答を生成
    def answer(self):
        if self.status != 0: return
        self.routeMode()
        self.stack = list()
        self.answerSub1()   # 道の探索
        self.answerSub2()   # GOAL処理

